'''
photo.py - draw growth curve using IRAF:PHOT
2008.08.29 YOSHIKAWA, Tomohiro
$Log: not supported by cvs2svn $
'''

import os, string
from astrocat import datalist
from pyraf import iraf
from mcsmdp import ds9
import pylab
import numpy

class GrowthCurve:
    def __init__(self, image):
        self.image = image

        self.fwhmpsf = None
        self.salgolism = None
        self.annulist = None

        self.photlist = None
        self.outbase = None

        self.markerlist = ['b--o', 'r--o', 'g--o', 'k--o', 'c--o', 'm--o', 'y--o',
                           'b--^', 'r--^', 'g--^', 'k--^', 'c--^', 'm--^', 'y--^',
                           'b--x', 'r--x', 'g--x', 'k--x', 'c--x', 'm--x', 'y--x']
        
        self.xpa = ds9.open()
        iraf.noao(_doprint=0)
        iraf.digiphot(_doprint=0)
        iraf.apphot(_doprint=0)
    def __del__(self):
        self._photdel_()

    def _photdel_(self):
        if self.photlist != None:
            for outfile in self.photlist:
                if os.access(outfile, os.F_OK): os.unlink(outfile)
        
    def setparam(self, fwhmpsf, salgolism='median',
                 annulist=[(4,2),(4,1),(4,0.5),(3,2),(3,1),(3,0.5)]):
        self.fwhmpsf = fwhmpsf
        self.salgolism = salgolism
        self.annulist = annulist
    
    def select(self, outbase=None):
        if self.outbase == None:
            self._photdel_()
        self.outbase = outbase
        tmp_coo = iraf.osfn(iraf.mktemp("tmp$photo_coo"))
        
        print "click the star on ds9."
        x, y = self.xpa.readcur()
        coo_f = open(tmp_coo, 'w')
        coo_f.write("%f %f\n"%(x,y))
        coo_f.close()
        
        index = 0
        self.photlist = []
        for annulus in self.annulist:
            if self.outbase == None:
                outfile = iraf.osfn(iraf.mktemp("tmp$photo_data"))
            else:
                outfile = "%s_%02d.dat"%(self.outbase,index)
            index += 1
            
            iraf.apphot.datapars.fwhmpsf=self.fwhmpsf
            apertures='%.4f:%.4f:%.4f'%(self.fwhmpsf/2., self.fwhmpsf*6, self.fwhmpsf/4.)
            iraf.phot(self.image, coords=tmp_coo, output=outfile, interactive=False,
                      verify=False, verbose=False,
                      calgorithm="centroid", cbox=5.,
                      annulus=self.fwhmpsf*annulus[0], dannulus=self.fwhmpsf*annulus[1],
                      apertures=apertures, salgorithm="median", sigma='INDEF',
                      datamin=-10000, datamax=10000,)
            self.photlist.append(outfile)
        if os.access(tmp_coo, os.F_OK): os.unlink(tmp_coo)
        
    def plot(self, imgname=None, sky=None, legend_loc='lower right'):
        if len(self.photlist) > len(self.markerlist):
            self.markerlist *= int(len(self.photlist) / len(self.markerlist)) + 1
        pylab.clf()
        index = 0
        for photfile,mark in zip(self.photlist, self.markerlist):
            rs = txdump(photfile, 'RAPERT')[0]
            an, dn = txdump(photfile, 'ANNULUS,DANNULUS')[0]
            flux = txdump(photfile, 'FLUX')[0]
            sky1 = txdump(photfile, 'MSKY')[0][0]
            # get flux at annulus
            #print rs
            #print an, dn
            for ri in range(len(rs)):
                if abs(rs[ri] - an) < 0.001: fa = flux[ri]
            #print index,an,dn,fa,sky1
            pylab.plot(rs, flux, mark,
                       label="%d: an=%.1f, dn=%.1f, flux@an=%.2f, sky=%.7f"%(index,an,dn,fa,sky1))
            index += 1

        if sky != None:
            al = []
            for an in self.annulist:
                a = an[0]
                try:
                    al.index(a)
                except ValueError:
                    al.append(a)
            al.sort()
            rs   = txdump(self.photlist[0], 'RAPERT')[0]
            sum  = txdump(self.photlist[0], 'SUM')[0]
            area = txdump(self.photlist[0], 'AREA')[0]
            flux = []
            for su, ar in zip(sum, area):
                flux.append(su - ar*sky)
            fa = []
            ra = []
            for an in al:
                for ri in range(len(rs)):
                    if abs(rs[ri] - an*self.fwhmpsf) < 0.001:
                        ra.append("%.1f"%(rs[ri]))
                        fa.append("%.2f"%(flux[ri]))
            
            pylab.plot(rs, flux, 'r-D', linewidth=2,
                       label="flux: %s (%s), sky=%.7f"%(string.join(fa), string.join(ra), sky))

        pylab.xlabel('radius [pix]')
        pylab.ylabel('flux')
        x, y = txdump(self.photlist[0], 'xcenter,ycenter')[0]
        pylab.title('growth curve: %s\nx=%.2f, y=%.2f'%(self.image, x, y))
        pylab.legend(loc=legend_loc, prop=pylab.matplotlib.font_manager.FontProperties(size='medium'))
        if imgname != None:
            pylab.savefig(imgname)

    def annulus(self, num, color='yellow'):
        x, y, an, dn = txdump(self.photlist[num], 'xcenter,ycenter,annulus,dannulus')[0]
        r2 = an + dn
        self.xpa.xpaset('regions', 'image;annulus(%s, %s, %s, %s)# color=%s move=0'%(x,y,an,r2,color))

    def print_flux(self, num, min=0, max=None):
        rs = txdump(self.photlist[num], 'rapert')[0]
        flux = txdump(self.photlist[num], 'flux')[0]
        if max == None: max = len(rs)
        sum = 0
        index = min
        for r, f in zip(rs[min:], flux[min:]):
            print "%d:\t%.1f\t%.2f"%(index,r,f)
            sum += f
            index += 1
            if index >= max: break
        print "average: %.2f"%(sum/(max-min))
    def flux(self, num):
        """
        returns a record of radius and flux specified by num 
        """
        rs = txdump(self.photlist[num], 'rapert')[0]
        flux = txdump(self.photlist[num], 'flux')[0]
        outlist = []
        for r,f in zip(rs,flux):
            outlist.append((r,f))
        return numpy.rec.array(outlist, names=['r', 'flux'])
    def sky(self, num, sum=False):
        """
        returns sky count specified by num
        sum: return the sum within the annulus 
        """
        sky = txdump(self.photlist[num], 'MSKY')[0][0]
        if sum:
            annulus = self.annulist[num]
            return sky * ( ((self.fwhmpsf*(annulus[0]+annulus[1]))**2 - (self.fwhmpsf*annulus[0])**2)* numpy.pi)
        else:
            return sky

def txdump(magfile, field, expr=True):
    if expr == True: expr = 'yes'
    lines = iraf.txdump(magfile, field, expr, Stdout=1)
    outlist = []
    for line in lines:
        tmplist = []
        for data in line.split():
            tmplist.append(datalist.str2data(data))
        outlist.append(tmplist)
    return outlist

def annulus(magfile, color='yellow'):
    """function annulus
    plot annulus on ds9
    magfile: output file of iraf:phot
    """
    xpa = ds9.open()
    iraf.digiphot()
    iraf.ptools()

    line = iraf.txdump(magfile, 'xcenter,ycenter,annulus,dannulus', 'yes', Stdout=1)
    x, y, an, dn = line[0].split()
    r2 = str(float(an) + float(dn))

    xpa.xpaset('regions', 'image;annulus(%s, %s, %s, %s)# color=%s move=0'%(x,y,an,r2,color))

## obsolete ##
def writelist(list, file):
    f_file = open(file, 'w')
    for data in list:
        f_file.write(data+'\n')
    f_file.close()

def growthcurve0(maglist):
    """function growthcurve
    draw growth curve from phot data file
    arguments
      maglist: output file of phot:iraf
    """
    pylab.clf()
    for magfile,mark in zip(maglist, ['b--o', 'r--o', 'g--o', 'k--o', 'c--o', 'm--o', 'y--o']):
        datalist = []
        for line in open(magfile):
            if line[0] == '#' or line[-3:-1] != '*\\': continue
            line=line.strip()
            data = line.split()
            datadic = []
            for k,v in zip(['rapert', 'sum', 'area', 'flux', 'mag', 'merr', 'pier', 'perror'], data):
                datadic.append((k,str2data(v)))
            datalist.append(dict(datadic))
        r = getarray(datalist, 'rapert')
        flux = getarray(datalist, 'flux')
        pylab.plot(r,flux, mark, label=magfile)
    pylab.xlabel('radius [pix]')
    pylab.ylabel('flux')
    pylab.title('growth curve')
    pylab.legend(loc='lower right')
    #return datalist
    #return r, flux
