# sfdb.py - database file for specfit
# $Log: not supported by cvs2svn $
# Revision 1.8  2008/06/04 09:39:35  tomohiro
# modify error message for SFDB
#
# Revision 1.7  2008/04/15 10:10:26  tomohiro
# modify bug to extract linetype parameter in sflogtab
#
# Revision 1.6  2008/03/19 10:32:55  tomohiro
# modify to accept line type with number
#
# Revision 1.5  2008/02/01 23:48:19  tomohiro
# add SFDB.get_comments()
# add function sflogtab()
#
# Revision 1.4  2007/09/24 09:06:17  tomohiro
# move radec to astrotool
#
# Revision 1.3  2007/09/03 08:45:51  tomohiro
# add SFLOG
#
# Revision 1.2  2007/07/17 08:54:32  tomohiro
# set database name automatically if None is given
#
# Revision 1.1  2007/07/17 02:37:27  tomohiro
# database file for specfit
#
import irafdb
import numpy
import pylab
import os,__builtin__
#try:
#    from rpy import *
#except ImportError:
#    print "failed to import rpy. Please install R and rpy, or you can't use specfit tools"
from scipy.stats import chi2
from numpy import rec
from astrocat import datalist
from iraftool import fileutil

## emission types
hatype = [('Halpha 6563', 'gaussian'),
          ('[NII] 6583', 'gaussian'),
          ('continuum', 'linear')]

o3type = [('[OIII] 5007', 'gaussian'),
          ('Hbeta 4861', 'gaussian'),
          ('continuum', 'linear')]

class SFDB(irafdb.IrafDB):
    def __init__(self, file=None, name=None):
        self.name = None
        if name is None and file is not None:
            name = os.path.basename(file)[2:]
        irafdb.IrafDB.__init__(self, file, name)
        self.__name__ = 'SFDB'
        if file == None:
            self['task'] = 'specfit'
        if not self.has_key('components'):
            self['components'] = []
        
    def writeto(self, file, mode='w'):
        ## set name
        name = os.path.basename(file)
        if name[:2] != 'sf':
            raise TypeError,"sf dababase file name should be sf*!"
        self.name = name[2:]
        keys = ['task', 'components'] + self.get_compnames()
        irafdb.IrafDB.writeto(self, file, mode, keys=keys)

    def get_compnames(self):
        outlist = []
        cnt = 1
        for c in self['components']:
            outlist.append(c[0]+str(cnt))
            cnt += 1
        return outlist

    def get_comments(self):
        outlist = []
        for comp in self.get_compnames():
            outlist.append(self.get_comment(comp))
        return outlist

    def add_component(self, comp):
        """
        add new component to SFDB
        comp: Component instance
        """
        self['components'].append([comp.name])
        cnt = len(self['components'])
        cname = comp.name+str(cnt)
        self[cname] = comp.tolist()
        if comp.comment != None:
            self.set_comment(cname, comp.comment)

    def del_component(self, key):
        """
        remove component
        key: component keyword name
        """
        ## check key name
        if not self.has_key(key):
            raise KeyError,"invalid key:",key
        cname = key[:-1]
        cnt   = int(key[-1])
        if self['components'][cnt-1][0] != cname:
            raise KeyError,"invalid key:",key

        ## copy old value and delete
        cmps = []
        coms = []
        for k in self.get_compnames():
            if k != key:
                cmps.append(self[k])
                coms.append(self.get_comment(k))
            del self[k]
        ## delete from components list
        del self['components'][cnt-1]
        
        ## renew components
        for k,cm,co in zip(self.get_compnames(), cmps, coms):
            self[k]=cm
            self.set_comment(k, co)
    
    def set_component(self, key, comp):
        """
        replace current component
        comp: Component instance to replace
        key:  component keyword name (component name + number)
        """
        if key[:-1] != comp.name: raise ValueError,"invalid key:",key
        self[key] = comp.tolist()
        self.set_comment(key, comp.comment)
    
    def get_component(self, name):
        complist = self[name]
        compcom  = self.get_comment(name)
        if name[:-1] == 'gaussian':
            comp = Gaussian(compcom)
        elif name[:-1] == 'linear':
            comp = Linear(compcom)
        else:
            raise ValueError, "component %s is not supported."%(name[:-1])
        for k,l in zip(comp.keys, complist):
            comp.set_param(k, l)
        return comp

    def plot(self, wmin, wmax, mode=None, keys=None, *args, **kargs):
        if keys == None:
            keys = self.get_compnames()
        if mode == None:
            mode = []
            for i in range(len(keys)):
                mode.append(None)
        w = numpy.arange(wmin, wmax)
        y = numpy.zeros(len(w))
        yy = []
        for k,m in zip(keys, mode):
            #print k
            comp = self.get_component(k)
            func = comp.get_func(m)
            y = y+ numpy.array(func(w))
            yy.append(numpy.array(func(w)))
        pylab.plot(w, y, *args, **kargs)

    def get_func(self, mode=None, keys=None):
        if keys == None:
            keys = self.get_compnames()
        if mode == None:
            mode = []
            for i in range(len(keys)):
                mode.append(None)
        funclist = []
        for k,m in zip(keys, mode):
            #print k
            comp = self.get_component(k)
            funclist.append(comp.get_func(m))
        
        def retfunc(x):
            y = 0
            for f in funclist:
                y += f(x)
            return y
        return retfunc

class SFLOG(SFDB):
    def __init__(self, file=None):
        SFDB.__init__(self, file, 'sflog')
    def _load(self, file, name):
        db_f = __builtin__.open(file)
        idx = 0
        aid = None
        for line in db_f.readlines():
            #print line
            if line[2:14] == 'Task SPECFIT':
                aid = idx
            idx += 1
        #print aid
        outdata = []
        db_f = __builtin__.open(file, 'r')
        if aid == None:
            raise ValueError,'SPECFIT record is not found'
        for i in range(aid):
            db_f.readline()
        # name
        while True:
            line = db_f.readline()
            if line[:9] == '** OBJECT': break
        data = line.split()
        self.name = fileutil.basename(data[4], dir=True, suffix=True)[0]
        # chisq
        while True:
            line = db_f.readline()
            if line[:9] == 'Chisquare': break
        data = line.split()
        chisq = float(data[2])
        pt    = int(data[4])
        npara = int(data[7])
        outdata.append(('chisq', (chisq, 'Chi square value')))
        outdata.append(('point', (pt, 'number of points')))
        outdata.append(('nparam', (npara, 'number of free parameters')))
        outdata.append(('rchisq', (chisq/(pt-npara), 'reduced Chi square value')))
        #outdata.append(('P', (r.pchisq(chisq, pt-npara, lower_tail=False), 'probability from Chi square distribution')))
        outdata.append(('P', (1 - chi2.cdf(chisq, pt-npara), 'probability from Chi square distribution')))

        # sample range
        while True:
            line = db_f.readline()
            if line[:6] == 'Sample': break
        data = line.split()
        wmin = float(data[2])
        wmax = float(data[4])
        outdata.append(('wmin', (wmin,'minimum sample range')))
        outdata.append(('wmax', (wmax, 'maximum sample range')))
        
        # components
        while True:
            line = db_f.readline()
            if line[:10] == 'components': break
        ncomp = int(line.split()[1])
        complist = []
        for i in range(ncomp):
            line = db_f.readline()
            complist.append([line.strip(),])
        outdata.append(('components', (complist, '')))
        
        while True:
            line = db_f.readline()
            if line[:11] == 'Param Value': break
        for i in range(ncomp):
            # components name
            line = db_f.readline()
            line, comment = line.split('#')[:2]
            comment = comment.strip()
            data = line.split()
            cname = data[0]
            vlen = int(data[1])
            # parameters
            data_matrix = []
            for j in range(vlen):
                line = db_f.readline()
                data_array = []
                for v in line.split():
                    data_array.append(datalist.str2data(v))
                data_matrix.append(data_array)
            outdata.append((cname, (data_matrix, comment)))
        db_f.close()
        return outdata

class Component:
    def __init__(self, name, keys, comment=None):
        self.name = name  # Component name
        self.keys = keys
        self.comment = comment
        self.data = {}
        for k in keys:
            self.data[k] = (None, None, None, None, None, None, None)
    
    def set_param(self, key, values):
        """
        set component parameter
        key: parameter keyword
        values: tuple of (value, lower, upper, step, tolerance, fix/free)
        """
        try:
            self.keys.index(key)
        except ValueError:
            raise ValueError, 'invalid key: '+key
        if len(values) != 6 and len(values) != 7:
            raise ValueError, "parameter list length shoud be 6 or 7"
        self.data[key] = tuple(values)
    
    def tolist(self):
        outlist = []
        for k in self.keys:
            outlist.append(list(self.data[k]))
        return outlist
    
    def get_func(self, mode=(0)):
        """
        get function to plot component
        mode:
          0: normal
          1: minimum parameters
          2: maximum parameters
        """
        raise NotImplementedError
    
    def plot(self, wmin, wmax, mode=None, *args, **kargs):
        """
        plot function using
        wmin, wmax: minimum/maximum wavelength
        *args, **kargs: arguments for plot
        """
        func = self.get_func(mode)
        w = numpy.arange(wmin, wmax)
        pylab.plot(w, func(w), *args, **kargs)

class Gaussian(Component):
    def __init__(self, comment=None):
        Component.__init__(self, 'gaussian', ['flux', 'centroid', 'fwhm', 'skew'], comment)

    def get_func(self, mode=(0,0,0,0)):
        #print mode
        if mode == None: mode = (0,0,0,0)
        if len(mode) != len(self.keys):
            raise ValueError, 'mode should be %d length tuple'%(len(self.keys))
        for m in mode:
            if m > 2:
                raise ValueError, 'mode should be less than 3'
        flux = self.data['flux'][mode[0]]
        cent = self.data['centroid'][mode[1]]
        fwhm = self.data['fwhm'][mode[2]]
        skew = self.data['skew'][mode[3]]

        if skew != 1:
            print "Warning: skew is not supported. set to 1"
            skew = 1
        fwhm_l = fwhm/2.998e5*cent
        s = fwhm_l/(2.*numpy.sqrt(2.*numpy.log(2.)))
        #print fwhm,fwhm_l,flux,s,cent
        return lambda x: flux*numpy.exp(-(x-cent)**2/2/s**2)/numpy.sqrt(2*numpy.pi)/s

class Linear(Component):
    def __init__(self, comment=None):
        Component.__init__(self, 'linear', ['flux', 'slope'], comment)

    def get_func(self, mode=(0,0)):
        #print mode
        if mode == None: mode = (0,0)
        if len(mode) != len(self.keys):
            raise ValueError, 'mode should be %d length tuple'%(len(self.keys))
        for m in mode:
            if m > 2:
                raise ValueError, 'mode should be less than 3'
        flux = self.data['flux'][mode[0]]
        slope = self.data['slope'][mode[1]]

        return lambda x: slope*(x - 1000) + flux

def open(file):
    return SFDB(file=file)

def sflogtab(infile, namelist=None, imfilelist=None, linetype=hatype):
    """
    sflogtab - make record table from log file of specfit
    infile - input file list [specfit*.log]
    namelist - list of object name
    imfilelist - list of 2D spec file list
    linetype - emission line name list for extract [list of (name, type)]
    """
    if namelist == None:
        namelist = ['N/A']*len(infile)
    if imfilelist == None:
        imfilelist = ['N/A']*len(infile)

    recarray = []
    for fl, name, imfile in zip(fileutil.fileparse(infile, base=False, check=True), namelist, imfilelist):
        sd = SFLOG(fl)
        sdlist = []
        for cname, ctype in zip(sd.get_comments(), sd.get_compnames()):
            sdlist.append((cname,ctype[:-1]))
        #print sdlist
        olinetype = []
        klist = []
        vlist = []
        for cname, ctype in linetype:
            #print cname, ctype
            #print sdlist.count((cname, ctype))
            if sdlist.count((cname, ctype)) != 1:
                if ctype == 'gaussian':
                    comp = Gaussian()
                elif ctype == 'linear':
                    comp = Linear()
                else:
                    raise ValueError, "component %s is not supported."%(ctype)
            else:
                cidx = sd.get_comments().index(cname)
                ck = sd.get_compnames()[cidx]
                comp = sd.get_component(ck)
            for k,v in comp.data.iteritems():
                key = cname.replace(' ', '-')+'_'+k
                #print k,v
                klist.append(key)
                vlist.append(v[0])
                klist.append(key+'_sig')
                vlist.append(v[1])
        for key in ['chisq','point','nparam','rchisq','P','wmin','wmax']:
            klist.append(key)
            vlist.append(sd[key])
        klist.append('NAME')
        vlist.append(name)
        klist.append('filename')
        vlist.append(imfile)
        recarray.append(vlist)
    return rec.array(recarray, names=klist)

def sflogrecord(inlog, linetype=hatype):
    """
    make specfit log file record
    inlog - input log file
    linetype - emission line name list for extract [list of (name, type)]
    """
    sd = SFLOG(inlog)
    sdlist = []
    for cname, ctype in zip(sd.get_comments(), sd.get_compnames()):
        sdlist.append((cname,ctype[:-1]))
    olinetype = []
    klist = []
    vlist = []
    for cname, ctype in linetype:
        if sdlist.count((cname, ctype)) != 1:
            if ctype == 'gaussian':
                comp = Gaussian()
            elif ctype == 'linear':
                comp = Linear()
            else:
                raise ValueError, "component %s is not supported."%(ctype)
        else:
            cidx = sd.get_comments().index(cname)
            ck = sd.get_compnames()[cidx]
            comp = sd.get_component(ck)
        for k,v in comp.data.iteritems():
            key = cname.replace(' ', '-')+'_'+k
            klist.append(key)
            vlist.append(v[0])
            klist.append(key+'_sig')
            vlist.append(v[1])
    for key in ['chisq','point','nparam','rchisq','P','wmin','wmax']:
        klist.append(key)
        vlist.append(sd[key])
    return klist, vlist
