"""
twodspec - extract and plot spectrum from 2-D spectrum FITS
2007.07.10. YOSHIKAWA, Tomohiro
$Log: not supported by cvs2svn $
Revision 1.8  2009/02/08 01:37:10  tomohiro
add bottom plot

Revision 1.7  2008/09/23 00:02:03  tomohiro
modify to omit emission labels out of wavelength limit in onedplot()

Revision 1.6  2008/02/01 23:56:13  tomohiro
add rest frame wavelength display

Revision 1.5  2007/10/07 05:41:30  tomohiro
add 2D spectrum plot option to onedplot

Revision 1.4  2007/09/26 12:05:31  tomohiro
add snplot, title, xlabel, ylabel, snstyle to onedplot()

Revision 1.3  2007/09/03 08:50:02  tomohiro
modify onedextract() to use noise extention image
add redshift and linelist parameter to onedplot()

Revision 1.2  2007/07/27 02:35:44  tomohiro
onedextract(): add noise calculation from background extention
onedplot(): add background plotting using BACKGROUND column

Revision 1.1  2007/07/17 07:22:07  tomohiro
extract and plot spectrum from 2-D spectrum FITS

"""
import pyfits,pylab,numpy,matplotlib
from matplotlib.ticker import IndexLocator, FuncFormatter, NullFormatter, MultipleLocator
from matplotlib import cm
from matplotlib.patches import Polygon
from pyraf import iraf
from iraftool import fileutil
from spectool import specdata,specline
try:
    import convolve
except ImportError:
    from stsci import convolve

c=2.99792458e18 # speed of light (A/s)
hp = 6.626e-27  # Planck's constant (erg s)

def onedextract(infile, line, center,
                outfile=None, nsum=5, width=1000,
                bgmin=10, bgmax=30, noise='noise',
                header=False
                ):
    fits_f = pyfits.open(fileutil.osimg(infile))
    data = fits_f[0].data
    head = fits_f[0].header
    data_bg = fits_f[1].data
    data_ns = fits_f[2].data
    crpix = head['CRPIX1']
    crval = head['CRVAL1']
    cdelt = head['CDELT1']
    #etime = head['EXPTIME']
    ttime = head['EXPTTIME']
    ncomb = head['NCOMBINE']
    #data /= etime
    #data_bg /= etime
    
    ymin = line - int(nsum/2.)
    ymax = line + int(nsum/2.)
    xcen = (center - crval)/cdelt + crpix
    xwid = width / cdelt
    xmin = int(xcen - xwid/2.)
    xmax = int(xcen + xwid/2.)
    if xmin <1: xmin=1
    if xmax >head['NAXIS1']: xmax=head['NAXIS1']
    pline=0
    #print ymin, ymax, xmin, xmax

    ## wavelength (A)
    wl = (numpy.arange(xmin, xmax+1) - crpix)*cdelt + crval
    
    ## sum of spectrum
    for dline in data[ymin-1:ymax, xmin-1:xmax]:
        pline += dline
    dnum = ymax - (ymin-1)
    
    bmin1 = line + int(bgmin/2.)
    bmax1 = line + int(bgmax/2.)
    bmin2 = line - int(bgmin/2.)
    bmax2 = line - int(bgmax/2.)

    ## average of background
    bline = 0
    #print bmin1, bmax1, bmin2, bmax2
    if noise == 'self':
        for dline in data[bmin1-1:bmax1, xmin-1:xmax]:
            bline += dline
        for dline in data[bmax2-1:bmin2:, xmin-1:xmax]:
            bline += dline
    else:
        for dline in data_bg[bmin1-1:bmax1, xmin-1:xmax]:
            bline += dline
        for dline in data_bg[bmax2-1:bmin2:, xmin-1:xmax]:
            bline += dline
    bgn = bmax1 - (bmin1-1) + bmin2 - (bmax2-1)
    bline /= bgn

    ## calcurate noise from background
    if noise == 'self':
        ## sample variance
        sig2=0
        for dline in data[bmin1-1:bmax1, xmin-1:xmax]:
            sig2 += (dline - bline)**2
        for dline in data[bmax2-1:bmin2:, xmin-1:xmax]:
            sig2 += (dline - bline)**2
        sig = numpy.sqrt(sig2 / bgn * dnum)
    elif noise == 'sky':
        ## Poisson distribution (with A-B subtruction)
        #return bline, dnum, hp, c, wl, ttime
        sig = numpy.sqrt(bline * dnum * (hp*c/wl**2) / ttime * 2.)
    else: # noise
        nline2 = 0
        for dline in data_ns[ymin-1:ymax, xmin-1:xmax]:
            nline2 += dline**2
        sig = numpy.sqrt(nline2 / ttime * 2.)
    
    head.update('MDP_NSUM', nsum, comment='# of pixels summed along the slit')
    head.update('MDP_SUMC', line, comment='center line for summation')
    head.update('MDP_IMSP', infile, comment='original 2D spec. file name')
    
    if outfile != None:
        specdata.mkspecdata(wl, pline, noise=sig, background=bline, outfile=outfile, header=head)
    else:
        if header:
            return specdata.mkspecdata(wl, pline, noise=sig, background=bline), head
        else:
            return specdata.mkspecdata(wl, pline, noise=sig, background=bline)

def onedplot(slist, style='b-', errorbar=True,
             xmin=None, xmax=None, ymin=None, ymax=None, yscale=1,
             boxcar=None, bin=None,
             ohfile="mdpdb$ohlist/ohlist_HK500.fits",
             rescurve="../STANDARD/StdMask2_HK500res.fits",
             error=True, snplot=False,
             fitfunc=None, grid=None, legend=False,
             title=None, xlabel='observed wavelength [A]', ylabel='erg/s/cm^s/A',
             ohstyle='k-', rstyle='k--', fstyle='k.', estyle='k:', lstyle='k-', snstyle='k-',
             redshift=0, restwl=True, linelist="mdpdb$linelist/nebulae.gaia", linewidth=10, linealpha=0.2,
             imspec=None, imymin=None, imymax=None, cnt=4, bottomplot=True
             ):
    """
    plot 1-D extracted spectrum data
    slist: specdata instance
    style: greph style for specdata
    errorbar: show error bars (True/False)
    ymin, ymax: min, max for yaxis of specdata
    boxcar: boxcar filter size for specdata (not apply if None)
    bin: bin size for specdata (not apply if None)
    ohfile: OH emission spectrum file ('self' for self background data, None for no display) 
    rescurve: response curve file (None for no display)
    error: display error (True/False)
    fitfunc: fitting function display with specdata
    grid: None: no display, 0: default grid, other numeric: ticks intervals
    legend: display legend for bottom graph (True/False)
    restwl: print rest-frame wavelength?
    bottomplot: bottom plot for reference data
    """
    # get data
    wl    = slist.field('WAVELENGTH')
    sdata = slist.field('FLUX')
    try:
        ndata = slist.field('NOISE')
    except NameError:
        ndata = None
        errorbar=False
        error=False
    if xmax == None:
        wmax = wl.max()
    else:
        wmax = xmax
    if xmin == None:
        wmin = wl.min()
    else:
        wmin = xmin

    if ohfile != None:
        if ohfile == 'self':
            ohdata = slist.field('BACKGROUND')
            ohwl = wl
        else:
            ohlist = pyfits.getdata(fileutil.osimg(ohfile))
            ohwl0  = ohlist.field('WAVELENGTH')
            ohwlm  = (ohwl0 > wmin) & (ohwl0 < wmax)
            ohdata = ohlist[ohwlm].field('FLUX')
            ohwl   = ohlist[ohwlm].field('WAVELENGTH')

    if rescurve != None:
        reslist = pyfits.getdata(fileutil.osimg(rescurve))
        reswl0  = reslist.field('WAVELENGTH')
        reswlm  = (reswl0 > wmin) & (reswl0 < wmax)
        resdata = reslist[reswlm].field('FLUX')
        reswl   = reslist[reswlm].field('WAVELENGTH')

    pylab.clf()
    ## define axes
    left, width = 0.15, 0.8
    if bottomplot:
        bottom, height = 0.05, 0.88
    else:
        bottom, height = 0.1, 0.8
    height_bottom, height_sn, height_img = 0.2, 0.2, 0.2

    axBottom = None
    axSN = None
    axImg = None
    # Bottom
    if bottomplot and (ohfile != None or rescurve != None or error):
        axBottom = pylab.axes([left, bottom, width, height_bottom])
        height -= height_bottom
        bottom += height_bottom
        axBottom.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%d'))
        if ohfile == None and rescurve == None: # error only
            axBottom.set_ylabel(ylabel)
            axBottom.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1e'))
        elif ohfile == None and not error: # rescurve only
            axBottom.set_ylabel('ADU/flux')
            axBottom.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1e'))
        elif rescurve == None and rescurve == None: # ohfile only
            axBottom.set_ylabel(ylabel)
            axBottom.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1e'))
        else:
            axBottom.yaxis.set_major_formatter(NullFormatter())
        #axBottom.yaxis.set_major_formatter(NullFormatter())
        axBottom.set_xlabel(xlabel)
    # SN plot
    if snplot:
        axSN = pylab.axes([left, bottom, width, height_sn])
        height -= height_sn
        bottom += height_sn
        if axBottom == None:
            axSN.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%d'))
        else:
            axSN.xaxis.set_major_formatter(NullFormatter())
    # spectrum image
    if imspec:
        if axBottom == None and axSN == None:
            axSpectrum = pylab.axes([left, bottom, width, height-height_img])
            bottom += height-height_img
            axImg = pylab.axes([left, bottom, width, height_img])
        else:
            axImg = pylab.axes([left, bottom, width, height_img])
            height -= height_img
            bottom += height_img
            axSpectrum = pylab.axes([left, bottom, width, height])
        axImg.yaxis.set_major_formatter(NullFormatter())
        axImg.xaxis.set_major_formatter(NullFormatter())
    else:
        # Spectrum
        axSpectrum = pylab.axes([left, bottom, width, height])
    if axBottom == None and axSN == None:
        axSpectrum.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%d'))
    else:
        axSpectrum.xaxis.set_major_formatter(NullFormatter())
    
    ## plot spectrum
    if boxcar != None:
        axSpectrum.plot(wl, convolve.boxcar(sdata, (boxcar,), mode="reflect")*yscale, style)
    elif bin != None:
        wl1=0
        fl1=0
        er1=0
        rem = len(sdata)%bin
        for i in range(bin):
            if rem > 0:
                wl1 += wl[:-1:bin]
                fl1 += sdata[:-1:bin]
                if errorbar: er1 += ndata[:-1:bin]**2
            else:
                wl1 += wl[::bin]
                fl1 += sdata[::bin]
                if errorbar: er1 += ndata[::bin]**2
            rem -= 1
        if errorbar:
            axSpectrum.errorbar(wl1/bin, fl1/bin*yscale, fmt=style, yerr=numpy.sqrt(er1)/bin)
        else:
            axSpectrum.plot(wl1/bin, fl1/bin*yscale, style)
        if snplot:
            axSN.plot(wl1[er1>0]/bin, (fl1[er1>0]/bin)/(numpy.sqrt(er1[er1>0])/bin), snstyle)
    else:
        if errorbar:
            axSpectrum.errorbar(wl, sdata*yscale, fmt=style, yerr=ndata)
        else:
            axSpectrum.plot(wl, sdata*yscale, style)
        if snplot:
            axSN.plot(wl, sdata/ndata, snstyle)
    if fitfunc != None:
        x = numpy.arange(wmin, wmax, 1.0)
        axSpectrum.plot(x, fitfunc(x)*yscale, fstyle)
    if grid != None:
        if grid > 0: axSpectrum.xaxis.set_ticks(wl[::grid])
        axSpectrum.grid()
        if snplot:
            if grid > 0: axSN.xaxis.set_ticks(wl[::grid])
            axSN.grid()
    if title != None:
        if axBottom == None and axSN == None and axImg != None:
            axImg.set_title(title, x=.5, y=1.10, fontsize='x-large')
        else:
            axSpectrum.set_title(title, x=.5, y=1.05)
    axSpectrum.set_xlim(wmin, wmax)
    if ymin != None:
        symin = ymin*yscale
    else:
        symin = None
    if ymax != None:
        symax = ymax*yscale
    else:
        symax = None
    axSpectrum.set_ylim(symin, symax)
    axSpectrum.set_ylabel(ylabel)
    if axBottom == None:
        axSpectrum.set_xlabel(xlabel)
    if snplot:
        axSN.set_xlim(wmin, wmax)
        axSN.set_ylabel('S/N')
    # plot linelist
    if linelist != None:
        linelist = iraf.osfn(linelist)
        vmin, vmax = axSpectrum.get_ylim()
        hmin, hmax = axSpectrum.get_xlim()
        v = pylab.arange(vmin, vmax+abs((vmax-vmin)*0.1), (vmax-vmin)*0.1)
        txp = vmax*0.9 + vmin*0.1
        #print redshift
        for em,ew in specline.line(z=redshift, linelist=linelist).iteritems():
            #print em,ew
            if ew > wmin and ew < wmax:
                axSpectrum.plot(pylab.ones(len(v))*ew, v, lstyle, linewidth=linewidth, alpha=linealpha)
                axSpectrum.text(ew, txp, em, rotation='vertical', ha='center', va='top', size=25)
        axSpectrum.set_xlim(hmin, hmax)
        axSpectrum.set_ylim(vmin, vmax)
    
    ## plot reference data
    if bottomplot:
        scalemin = None
        scalemax = None
        if ohfile != None:
            axBottom.plot(ohwl, ohdata, ohstyle, label='OH emission')
            scalemin = ohdata.min()
            ohdata.sort()
            scalemax = ohdata[int(len(ohdata)*.96)]
            #print scalemin, scalemax
        
        if error:
            if scalemin != None:
                sc = (scalemax - scalemin) / (ndata.max() - ndata.min())
                axBottom.plot(wl, sc*(ndata-ndata.min()) + scalemin, estyle, label='error bar')
            else:
                axBottom.plot(wl, ndata, estyle, label='error bar')
                scalemin = ndata.min()
                scalemax = ndata.max()
        if rescurve != None:
            if scalemin != None:
                # scale rescurve
                sc = (scalemax - scalemin) / (resdata.max() - 0)
                axBottom.plot(reswl, sc*(resdata) + scalemin, rstyle, label='response curve')
            else:
                axBottom.plot(reswl, resdata, rstyle, label='response curve')
                scalemin = 0

        if ohfile != None or rescurve != None or error:
            axBottom.set_xlim(wmin, wmax)
            axBottom.set_ylim(scalemin, scalemax)
            if legend: axBottom.legend()
            if grid != None:
                if grid > 0: axBottom.xaxis.set_ticks(wl[::grid])
                axBottom.grid()
    else:
        if error:
            f_error = [(wmin, 0)]+zip(wl, ndata*yscale)+[(wmax, 0)]
        fill = Polygon(f_error, facecolor='#bbbbbb', edgecolor='k', linewidth=0)
        axSpectrum.add_patch(fill)
    
    if redshift and restwl:
        xtloc = axSpectrum.get_xticks()
        y = axSpectrum.get_ylim()[1]
        for x in xtloc[1:-1]:
            axSpectrum.text(x, y, '%d'%(x/(redshift+1.)), ha='center', va='bottom')
        x = numpy.array(axSpectrum.get_xlim()).mean()
        axSpectrum.text(x, y*1.02, 'rest frame wavelength [A]', ha='center', va='bottom')
    
    if axImg != None:
        imdata,head = pyfits.getdata(fileutil.osimg(imspec), 0, header=True)
        crpix = head['CRPIX1']
        crval = head['CRVAL1']
        cdelt = head['CDELT1']
        imxmin = int((wmin - crval)/cdelt + crpix)
        imxmax = int((wmax - crval)/cdelt + crpix)
        nsum = int(imymax - imymin)
        
        if ymin == None:
            ymin = axSpectrum.get_ylim()[0]
        zmin = ymin/nsum*cnt/yscale
        #zmin = ymin/4
        if ymax == None:
            ymax = axSpectrum.get_ylim()[1]
        zmax = ymax/nsum*cnt/yscale
        #zmax = ymax/4
        print 'zmin=%e, zmax=%e'%(zmin, zmax)
        axImg.imshow(imdata[imymin:imymax,imxmin:imxmax], cmap=cm.gist_heat, origin='lower', interpolation='nearest',
                     vmin=zmin, vmax=zmax, aspect='auto')

    #pylab.show()
    pylab.draw()

def onedplot2(ax, slist,
              style='b-', xmin=None, xmax=None, xscale=1,
              ymin=None, ymax=None, yscale=1,
              boxcar=None, bin=None, grid=None, ticks=None,
              errorbar=True, error=True, fitfunc=None, fstyle='k.', 
              title=None,
              xlabel=r'$\lambda_{\rm obs} \left[\AA\right]$',
              ylabel=r'$F_\lambda\,[\times10^{-18}\,erg\,cm^{-2}s^{-1}\AA^{-1}]$',
              linelist="mdpdb$linelist/nebulae.gaia", redshift=0,
              linewidth=10, linealpha=0.2, lstyle='k-', ltext=True,
              ):
    """
    plot spectram in an axis
    
    ** parameters **
      ax: axis to plot
    ** data **
     slist:  spectrum data
    ** plot **
     style:  plot style
     xmin, xmax, ymin, ymax: axis limits for plotting
     yscale: scale for spectrum data
     boxcar: apply boxcar filter in plotting
     bin:    apply binning in plotting
     grid:   show grid
     ticks:  interval of ticks
    ** reference data **
     errorbar: show error bars
     error:    plot error in grey region
     fitfunc:  plot best-fit function
     fstyle:   line style for best-fit function
    ** labels and title **
     title:    title string
     xlabel:   x-label string
     ylabel:   y-label string
    ** line list **
     linelist:  line list file name
     redshift:  redshift of the object
     linewidth: line width for line list
     linealpha: line alpha for line list
     lstyle:    line stile for line list
     ltext:     show line name in a text
    """
    # get data
    wl    = slist.field('WAVELENGTH')
    sdata = slist.field('FLUX')
    try:
        ndata = slist.field('NOISE')
    except NameError:
        ndata = None
        errorbar=False
        error=False
    if xmax == None:
        wmax = wl.max()
    else:
        wmax = xmax
    if xmin == None:
        wmin = wl.min()
    else:
        wmin = xmin
    imin, imax = pylab.searchsorted(wl, (wmin, wmax))
    if ymax == None:
        fmax = sdata[imin:imax].max()
    else:
        fmax = ymax
    if ymin == None:
        fmin = sdata[imin:imax].min()
    else:
        fmin = ymin
    
    ## plot spectrum
    if boxcar != None:
        ax.plot(wl*xscale, convolve.boxcar(sdata, (boxcar,), mode="reflect")*yscale, style)
    elif bin != None:
        wl1=0
        fl1=0
        er1=0
        rem = len(sdata)%bin
        for i in range(bin):
            if rem > 0:
                wl1 += wl[:-1:bin]
                fl1 += sdata[:-1:bin]
                if errorbar: er1 += ndata[:-1:bin]**2
            else:
                wl1 += wl[::bin]
                fl1 += sdata[::bin]
                if errorbar: er1 += ndata[::bin]**2
            rem -= 1
        if errorbar:
            ax.errorbar(wl1/bin*xscale, fl1/bin*yscale, fmt=style, yerr=numpy.sqrt(er1)/bin)
        else:
            ax.plot(wl1/bin*xscale, fl1/bin*yscale, style)
    else:
        if errorbar:
            ax.errorbar(wl*xscale, sdata*yscale, fmt=style, yerr=ndata*yscale)
        else:
            ax.plot(wl*xscale, sdata*yscale, style)
    ## best-fit function
    if fitfunc != None:
        x = numpy.arange(wmin, wmax, 1.0)
        ax.plot(x*xscale, fitfunc(x)*yscale, fstyle)
        fmin = fitfunc(x).min()
        fmax = fitfunc(x).max()
    ## grid
    if grid != None:
        if grid > 0: ax.xaxis.set_ticks(wl[::grid]*xscale)
        ax.grid()
    ## ticks
    if ticks > 0:
        ax.xaxis.set_ticks(pylab.arange(pylab.ceil(xmin/ticks)*ticks, pylab.ceil(xmax/ticks)*ticks, ticks)*xscale)
    ## title and labels
    if title != None:
        #ax.set_title(title, x=.5, y=1.05)
        ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    ## limits of axes
    ax.set_xlim(wmin*xscale, wmax*xscale)
    if ymin != None:
        symin = ymin*yscale
    else:
        symin = (1.2*fmin - 0.2*fmax)*yscale
    if ymax != None:
        symax = ymax*yscale
    else:
        symax = (1.2*fmax - 0.2*fmin)*yscale
    #print symin, symax
    ax.set_ylim(symin, symax)

    ## adjust yticks
    ax.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1f'))
    yt = ax.yaxis.get_ticklocs()
    if len(yt) > 3:
        it = int(len(yt)/3.)
        print yt, it
        ax.yaxis.set_ticks(yt[::it])

    ## error
    if error:
        f_error = [(wmin*xscale, 0)]+zip(wl*xscale, ndata*yscale)+[(wmax*xscale, 0)]
        fill = Polygon(f_error, facecolor='#bbbbbb', edgecolor='k', linewidth=0)
        ax.add_patch(fill)

    ## line list
    if linelist != None:
        linelist = iraf.osfn(linelist)
        vmin, vmax = ax.get_ylim()
        hmin, hmax = ax.get_xlim()
        v = pylab.arange(vmin, vmax+abs((vmax-vmin)*0.1), (vmax-vmin)*0.1)
        txp = vmax*0.9 + vmin*0.1
        #print redshift
        for em,ew in specline.line(z=redshift, linelist=linelist).iteritems():
            #print em,ew
            if ew > wmin and ew < wmax:
                ax.plot(pylab.ones(len(v))*ew*xscale, v, lstyle, linewidth=linewidth, alpha=linealpha)
                if ltext: ax.text(ew*xscale, txp, em, rotation='vertical', ha='center', va='top', size=25)
        ax.set_xlim(hmin, hmax)
        ax.set_ylim(vmin, vmax)

